import collections
import copy
import hashlib
import json
import re
import time
from logging import info, warning

import act.api
from act.api import RE_UUID_MATCH

from . import DEFAULT_VALIDATOR
from .base import ActBase, Comment, NameSpace, Organization, Origin, origin_lookup_serializer
from .obj import Object, ObjectType
from .schema import Field, MissingField, ValidationError, schema_doc


class IllegalFactChain(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class RelevantObjectBindings(ActBase):
    SCHEMA = [
        Field(
            "source_object_type",
            deserializer=ObjectType,
            serializer=lambda source_object_type: source_object_type.id),
        Field(
            "destination_object_type",
            deserializer=ObjectType,
            serializer=lambda destination_object_type: destination_object_type.id),
        Field(
            "bidirectional_binding",
            default=False)]

    def __hash__(self):
        """
        Hash of the object binding is the combination of
        source/destination object and bidirectional
        """

        if self.source_object_type:
            source = self.source_object_type.id
        else:
            source = None

        if self.destination_object_type:
            destination = self.destination_object_type.id
        else:
            destination = None

        return hash((source, destination, self.bidirectional_binding))

    def __eq__(self, other):
        "Equality operator"

        return hash(self) == hash(other)


class RelevantFactBindings(ActBase):
    """Meta Fact Type"""
    SCHEMA = [
        Field("name"),
        Field("id"),
        Field("fact_type", flatten=True),
    ]

    def __hash__(self):
        "Hash of the fact bindings is the ID itself"

        return hash(self.id)

    def __eq__(self, other):
        "Equality operator"

        return hash(self) == hash(other)

    def serialize(self):
        # Return None for empty objects (non initialized objects)
        if not self.id:
            return None

        return {"factType": self.id}

    def __bool__(self):
        # Return None for empty objects (non initialized objects)
        if not self.id:
            return False

        return True


class FactType(ActBase):
    """Manage FactType"""
    SCHEMA = [
        Field("name"),
        Field("id"),
        Field("default_confidence"),
        Field("validator", default="RegexValidator"),
        Field("validator_parameter", default=DEFAULT_VALIDATOR),
        Field("relevant_object_bindings", deserializer=RelevantObjectBindings),
        Field("relevant_fact_bindings", deserializer=RelevantFactBindings),
        Field("namespace", deserializer=NameSpace),
    ]

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(FactType, self).__init__(*args, **kwargs)

    def add(self):
        params = self.serialize()

        fact_type = self.api_post("v1/factType", **params)["data"]

        # Empty data and load new result from response
        self.data = {}
        self.deserialize(**fact_type)

        info("Created fact type: {}".format(self.name))

        return self

    def add_object_binding(
            self,
            source_object_type,
            destination_object_type,
            bidirectional_binding=False):
        """Add bindings
Args:
    source_object_type (objectType):        Source Object Type
    destination_object_type (objectType):   Destination Object Type
    bidirectional_binding (boolean):        Whether the binding is bidirectional
"""
        return self.add_object_bindings([
            RelevantObjectBindings(
                source_object_type,
                destination_object_type,
                bidirectional_binding)])

    def add_object_bindings(self, add_bindings):
        """Add multiple object bindings
Args:
    add_bindings (relevantObjectBindings[]): List of bindings which must be a
                                             tuple of RelevantObjectBindings
"""
        if not self.id:
            raise MissingField("Must have fact type ID")

        url = "v1/factType/uuid/{}".format(self.id)

        existing_bindings = set(self.relevant_object_bindings)

        # Exclude already existing bindings
        new_bindings = [
            binding
            for binding
            in add_bindings
            if binding not in existing_bindings]

        if not new_bindings:
            warning(
                "All bindings specified for {} already exists".format(
                    self.name))
            return self

        serialized_bindings = [binding.serialize() for binding in new_bindings]

        fact_type = self.api_put(
            url, addObjectBindings=serialized_bindings)["data"]

        self.data = {}
        self.deserialize(**fact_type)

        # Emit log of created bindings
        for binding in new_bindings:
            info(
                "Added bindings to fact type {}: {} -> {} (bidirectional_binding={})".format(
                    self.name, binding.source_object_type.data.get(
                        "name", None), binding.destination_object_type.data.get(
                            "name", None), binding.bidirectional_binding))

        return self

    def add_fact_binding(self, fact_type):
        """Add bindings
Args:
    fact_type (factType):   Fact type
"""
        return self.add_fact_bindings([RelevantFactBindings(name=fact_type.name, id=fact_type.id)])

    def add_fact_bindings(self, add_bindings):
        """Add multiple fact bindings
Args:
    add_bindings (relevantFactBindings[]): List of Fact bindings
"""
        if not self.id:
            raise MissingField("Must have fact type ID")

        url = "v1/factType/uuid/{}".format(self.id)

        existing_bindings = set(self.relevant_fact_bindings)

        # Exclude already existing bindings
        new_bindings = [
            binding
            for binding
            in add_bindings
            if binding not in existing_bindings]

        if not new_bindings:
            warning("All bindings specified for {} already exists".format(self.name))
            return self

        serialized_bindings = [binding.serialize() for binding in new_bindings]

        fact_type = self.api_put(
            url, addFactBindings=serialized_bindings)["data"]

        self.data = {}
        self.deserialize(**fact_type)

        # Emit log of created bindings
        for binding in new_bindings:
            info("Added bindings to meta fact type {}: {}".format(self.name, binding.name))
        return self

    def rename(self, name):
        if not self.id:
            raise MissingField("Must have fact type ID")

        old_name = self.name

        url = "v1/factType/uuid/{}".format(self.id)
        fact_type = self.api_put(url, name=name)["data"]

        self.data = {}
        self.deserialize(**fact_type)

        info("Renamed fact type {}: {} -> {}".format(self.id, old_name, name))

        return self


class ReferencedFact(ActBase):
    """Retracted Fact"""

    SCHEMA = [
        Field("type", deserializer=FactType,
              serializer=lambda fact_type: fact_type.name),
        Field("value", default=""),
        Field("id")
    ]

    def __eq__(self, other):
        """ Check equality with other object """

        # If other is None, return True if id, type and value is None
        if other is None:
            if not (self.id or self.type.name or self.value):
                return True
            return False

        # Otherwise, use equality check from super class
        return super(ReferencedFact, self).__eq__(other)

    def serialize(self):
        # Return None for empty objects (non initialized objects)
        if not (self.id or self.type.name):
            return None
        # return default serializer
        return super(ReferencedFact, self).serialize()

    def __bool__(self):
        # Return None for empty objects (non initialized objects)
        if (self.type.name or self.value or self.id):
            return True
        return False


def object_serializer(obj):
    if not obj:
        return None

    if obj.id is None and obj.type.name is None and obj.value is None:
        return None

    if obj.id:
        return obj.id

    return "{}/{}".format(obj.type.name, obj.value)


class Fact(ActBase):
    """Manage facts"""

    SCHEMA = [
        Field("type", deserializer=FactType,
              serializer=lambda fact_type: fact_type.name),
        Field("value", default=""),
        Field("id", serializer=False),
        Field("flags", serializer=False),
        Field("origin", deserializer=Origin),
        Field("comment"),
        Field("added_by", deserializer=Origin, serializer=False),
        Field("trust", serializer=False),
        Field("confidence"),
        Field("certainty", serializer=False),
        Field("timestamp", serializer=False),
        Field("last_seen_timestamp", serializer=False),
        Field("in_reference_to", deserializer=ReferencedFact),
        Field("organization", deserializer=Organization),
        Field("access_mode", default="Public"),
        Field("source_object", deserializer=Object),
        Field("destination_object", deserializer=Object),
        Field("bidirectional_binding", default=False),
    ]

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(Fact, self).__init__(*args, **kwargs)

    def __add_if_not_exists(self, *args, **kwargs):
        """Add binding if it does not exist"""
        binding = Object(*args, **kwargs)
        if binding not in self.data["objects"]:
            self.data["objects"].append(binding)

        return self

    def source(self, *args, **kwargs):
        """Add source binding. Accepts the same arguments as Object()"""

        self.data["source_object"] = Object(*args, **kwargs)

        if not self.source_object.id and not (
                self.source_object.type and self.source_object.value):
            raise MissingField(
                "Must have either object_id or object_type and object_value")

        return self

    def destination(self, *args, **kwargs):
        """Add source binding. Accepts the same arguments as Object()"""

        self.data["destination_object"] = Object(*args, **kwargs)

        if not self.destination_object.id and not (
                self.destination_object.type and self.destination_object.value):
            raise MissingField(
                "Must have either object_id or object_type and object_value")

        return self

    def bidirectional(
            self,
            source_object_type,
            source_object_value,
            destination_object_type,
            destination_object_value):
        """Add bidirectional binding."""

        self.data["source_object"] = Object(
            source_object_type,
            source_object_value)
        self.data["destination_object"] = Object(
            destination_object_type,
            destination_object_value)
        self.data["bidirectional_binding"] = True

        return self

    def get(self):
        """Get fact"""

        if not self.id:
            raise MissingField(
                "Must have fact ID to get, or use search instead")

        fact = self.api_get("v1/fact/uuid/{}".format(self.id))["data"]
        self.data = {}
        self.deserialize(**fact)
        return self

    def add(self):
        """Add (meta) fact
Add this fact to the platform.

Reutrns the newly created fact.
"""
        if self.in_reference_to:
            # This is a meta fact
            return self.__add_meta()

        # This is not a meta fact
        return self.__add_fact()

    def __add_fact(self):
        """Add fact

This is not called directly, but are called from add() if Fact is not a meta fact.
        """
        started = time.time()

        # Construct parameters to be sent to backend when creating fact
        params = {
            k: v
            for k, v in self.serialize().items()
            # Exclude inReferenceTo, which is added automatically for retracted facts and meta facts
            if k not in ("inReferenceTo")
        }

        # Special serializer for source/destination objects
        params["sourceObject"] = object_serializer(self.source_object)
        params["destinationObject"] = object_serializer(self.destination_object)

        # Special serializer for origin
        params["origin"] = origin_lookup_serializer(self.origin, config=self.config)

        fact = self.api_post("v1/fact", **params)["data"]

        # Empty data and load new result from response
        self.data = {}
        self.deserialize(**fact)

        info("Created fact in %.2fs: data=%s" % (time.time() - started, json.dumps(fact)))

        return self

    def __add_meta(self):
        """Add meta fact to platform

This is not called directly, but are called from add() if Fact is a meta fact.
"""
        started = time.time()
        if not self.in_reference_to:
            raise MissingField("Fact is not a meta fact (in_reference_to is not set)")

        if not self.in_reference_to.id:
            raise MissingField("Referenced fact must have fact ID")

        params = {
            k: v
            for k, v in self.serialize().items()
            if k not in ("inReferenceTo", "bidirectionalBinding") and v
        }

        url = "v1/fact/uuid/{}/meta".format(self.in_reference_to.id)
        meta_fact = self.api_post(url, **params)["data"]

        self.data = {}
        self.deserialize(**meta_fact)
        info("Created meta fact in %.2fs: data=%s" % (time.time() - started, json.dumps(meta_fact)))

        return self

    # pylint: disable=unused-argument,dangerous-default-value
    def meta(self, *args, **kwargs):
        """Create meta fact
Creates a new meta fact with reference to this fact.
Takes the same arguments as Fact(), except for in_reference_to.
Returns meta fact
    """

        ref = ReferencedFact(type=self.type, value=self.value, id=self.id)

        meta = Fact(*args, in_reference_to=ref, **kwargs)

        # Add config to meta fact (user/auth)
        meta.configure(self.config)

        return meta

    # pylint: disable=unused-argument,dangerous-default-value

    def get_acl(self):
        """Get acl"""

        if not self.id:
            raise MissingField("Must have fact ID to resolve ACL")

        return self.api_get("v1/fact/uuid/{}/access".format(self.id))["data"]

    def grant_access(self, subject_uuid):
        """Grant access - not implemented yet"""

        raise act.api.base.NotImplemented(
            "Grant access is not implemented, ignoring {}".format(subject_uuid))

    def get_comments(self):
        """Get comments
Args:
    None

Returns ActResultSet of Comments.

        """

        if not self.id:
            raise MissingField("Must have fact ID to get comments")

        res = self.api_get("v1/fact/uuid/{}/comments".format(self.id))
        return act.api.base.ActResultSet(res, Comment)

    def add_comment(self, comment, reply_to=None):
        """Add comment
Args:
    comment (str):    Comment to add
    reply_to (uuid):  Comment is a reply to another comment (specified by uuid)

Returns Fact object
        """

        if not self.id:
            raise MissingField("Must have fact ID to add comment")

        if reply_to and not re.search(RE_UUID_MATCH, reply_to):
            raise ValidationError(
                "reply_to is not a valid UUID: {}".format(reply_to))

        self.api_post(
            "v1/fact/uuid/{}/comments".format(self.id),
            comment=comment, replyTo=reply_to)

        return self

    # pylint: disable=unused-argument,dangerous-default-value
    def get_meta(
            self,
            before=None,
            after=None,
            limit=None):
        """Get meta facts
Args:
    before (timestamp):           Only return Facts added before a specific
                                  timestamp. Timestamp is on this format:
                                  2016-09-28T21:26:22Z
    after (timestamp):            Only return Facts added after a specific
                                  timestamp. Timestamp is on this format:
                                  2016-09-28T21:26:22Z
    limit (integer):              Limit the number of returned Objects
                                  (default 25, 0 means all)
    """

        params = act.api.utils.prepare_params(locals())

        if not self.id:
            raise MissingField("Must have fact ID to get comments")

        res = self.api_get("v1/fact/uuid/{}/meta".format(self.id), params=params)
        return act.api.base.ActResultSet(res, Fact)

    def retract(
            self,
            organization=None,
            source=None,
            access_mode=None,
            comment=None,
            acl=[]):
        """Retract fact
Args:
    organization (str): Set owner of new Fact. If not set the current user's
                        organization will be used (takes Organization UUID)
    source (str):       Set Source of new Fact. If not set the current user
                        will be used as Source (takes Source UUID)
    access_mode:        Set access mode of new Fact. If not set the accessMode
                        from the retracted Fact will be used = ['Public',
                        'RoleBased', 'Explicit']
    comment (str):      If set adds a comment to new Fact
    acl (str[] | str):  If set defines explicitly who has access to new Fact (takes Subject UUIDs)

All arguments are optional.

Returns retracted fact.
    """

        params = act.api.utils.prepare_params(locals(), ensure_list=["acl"])

        if self.id:
            url = "v1/fact/uuid/{}/retract".format(self.id)
        else:
            raise MissingField(
                "Must have object ID to retract object")

        fact = self.api_post(url, **params)["data"]

        self.data = {}
        self.deserialize(**fact)

        return self

    def __hash__(self):
        """
        Hash of the fact. We use the string representation of the fact as seed.
        """

        return hash(str(self))

    def __str__(self):
        """
        Construnct string representation on this format
        (src_obj_type/src_obj_value) -[fact_type/fact_value]-> (dest_obj_type/dest_obj_value)
        """

        out = ""

        # Include source object if set
        if self.source_object:
            out += "({}/{}) -".format(self.source_object.type.name, self.source_object.value)

        # Add fact type
        out += "[{}".format(self.type.name)

        # Add value if set
        if self.value and not self.value.startswith("-"):
            out += "/{}".format(self.value)
        out += "]"

        # Add destination object if set
        if self.destination_object:

            # Use arrow unless bidirectional
            if self.bidirectional_binding:
                out += "-"
            else:
                out += "->"

            out += " ({}/{})".format(self.destination_object.type.name,
                                     self.destination_object.value)

        return out


def fact_chain_seed(*facts):
    """
    Get seed of the fact chain. The seed is the string representations of all facts sorted
    and joined with newline.

Args:
    *facts [Facts]: ACT facts

Returns: string representation of the seed
    """

    known_objects = [fact.source_object
                     for fact in facts
                     if fact.source_object.value != "*"] + \
                    [fact.destination_object
                     for fact in facts
                     if fact.destination_object.value != "*"]

    if len(known_objects) != 2:
        raise IllegalFactChain(
            "There should be exactly two known objects in a fact chain: {}".format(facts))

    return "\n".join(sorted([str(fact) for fact in facts]))


def fact_chain(*facts):
    """
    Return fact chain of all facts.
    A fact chain is a modified version of the facts, where all placeholder
    object values are replace with a hash using the all incoming/outgoing
    links from the placeholder as a seed

Args:
    *facts [Facts]: ACT facts

Returns: facts [Facts]: ACT facts
    """

    # Calculate the hash for this chain, based on the seed
    sha256sum = hashlib.sha256(fact_chain_seed(*facts).encode("utf8")).hexdigest()

    chain = copy.copy(facts)

    # Return the fact chain, with all object values replaced "[placeholder[<HASH>]]"
    for fact in chain:
        for obj in (fact.source_object, fact.destination_object):
            if obj.value == "*":
                obj.data["value"] = "[placeholder[{}]]".format(sha256sum)

    return chain



import copy
import json
from .utils import snake_to_camel, camel_to_snake
from logging import info, warning

def default_deserializer(value):
    """Default serializer. return trimmed value"""

    if isinstance(value, str):
        trimmed = value.strip()

        if trimmed != value:
            info('Value was trimmed: "{}"'.format(value))
            value = trimmed

    return value


class ValidationError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class MissingField(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


# pylint: disable=useless-object-inheritance
class Field(object):
    """Schema fields"""

    def __init__(
            self,
            name,
            default=None,
            serializer=None,
            deserializer=default_deserializer,
            serialize_target=None,
            deserialize_target=None,
            flatten=False):
        """
Args:
    name (str):                Field name
    default (str):             Default value for field if not specified
    serializer (func|False):   Function used to serialize the field. Set to
                               False if field should not be serialized.
    deserializer (func|False): Function used to deserialize data. Set to
                               False if field should not be deserialized.
    serialize_target (str):    Move data to this key(name) when erializing
    deserialize_target (str):  Move data to this key(name) when deserializing
    flatten (bool):            Flatten data structure
    """
        self.name = name
        self.default = default
        self.serializer = serializer
        self.deserializer = deserializer
        self.serialize_target = serialize_target
        self.deserialize_target = deserialize_target
        self.flatten = flatten

    def deserialize(self, value):
        """Deserialize field using suplied value"""

        # If value is already a Schema object, use it as is
        if isinstance(value, Schema):
            return value

        if isinstance(value, dict):
            # pylint: disable=comparison-with-callable
            if self.deserializer == default_deserializer:
                raise ValidationError(
                    "dict is not supported by default serializer. field={}, value={}".format(
                        self.name, value))

            return self.deserializer(**value)

        return self.deserializer(value)

    def serialize(self, value):
        """Serialize field using suplied value"""

        # Custom serializer
        if self.serializer:
            return self.serializer(value)

        # Nested serializer
        if hasattr(value, "serialize") and callable(value.serialize):
            return value.serialize()

        return value


def schema_doc(schema):
    """Dynamic create doctstring from schema fields"""
    def dec(obj):
        docstring = obj.__doc__ or ""
        dynamic = """

The following arguments can be specified by position or keyword.
    {}
""".format(str("\n    ".join([field.name for field in schema])))

        obj.__doc__ = docstring + dynamic

        return obj
    return dec


# pylint: disable=useless-object-inheritance
class Schema(object):
    """Define schema and serialize/deserialize response from the platform"""

    def __init__(self, *args, **kwargs):
        """Initialize (deserialize) fields"""

        self.data = {}

        # Apply all positinal arguments based on order in SCHEMA
        for i, v in enumerate(args):
            kwargs[self.SCHEMA[i].name] = v

        self.deserialize(**kwargs)

    def json(self, exclude_none=True, to_camel_case=True):
        return json.dumps(
            self.serialize(
                exclude_none=exclude_none,
                to_camel_case=to_camel_case))

    def serialize(self, exclude_none=True, to_camel_case=True):
        entries = {}
        for field in self.SCHEMA:
            serialize_target = field.serialize_target or field.name

            if to_camel_case:
                serialize_target = snake_to_camel(serialize_target)

            # Entry is stored in another field internally
            value = self.data.get(field.name, None)

            # Exclude empty values
            if exclude_none and value is None:
                continue

            # Serializer disabled
            if field.serializer is False:
                continue

            if isinstance(value, (list, tuple)):
                entries[serialize_target] = [field.serialize(v) for v in value]
            else:
                entries[serialize_target] = field.serialize(value)

        return entries

    def get_field(self, name):
        for field in self.SCHEMA:
            if name == field.name:
                return field
        return None

    def get_deserialize_field(self, name):
        field = self.get_field(name)

        if not field:
            return None

        if field.deserialize_target:
            return self.get_field(field.deserialize_target)
        return field

    def get_serialize_field(self, name):
        field = self.get_field(name)

        if not field:
            return None

        if field.serialize_target:
            return self.get_field(field.serialize_target)

        return field

    def deserialize(self, **entries):
        for k, value in entries.items():
            k = camel_to_snake(k)
            field = self.get_deserialize_field(k)

            if not field:
                warning('"{}" not defined in schema on {}'.format(k, self.__class__))
                continue

            # deserializer disabled
            if field.deserializer is False:
                continue

            if field.flatten:
                self.deserialize(**value)
                continue

            if isinstance(value, (list, tuple)):
                self.data[field.name] = [field.deserialize(v) for v in value]
            else:
                self.data[field.name] = field.deserialize(value)

        if not hasattr(self, "SCHEMA"):
            raise ValidationError(
                "No SCHEMA defined in class {}".format(self.__class__))

        for field in self.SCHEMA:
            # Loop through all fields and set default value, unless any of
            # - field is deseriazlied with another key
            # - field is flattened
            # - deserializer is disabled
            if field.name not in self.data \
                    and not field.deserialize_target \
                    and not field.flatten \
                    and field.deserializer is not False:
                self.data[field.name] = copy.copy(field.default)

    def __getitem__(self, key):
        return self.data[key]

    def __getattr__(self, attr):
        """
        Get attribute from schema
        """
        if attr in self.__dict__.get("data", {}):
            return self.__dict__["data"][attr]

        raise AttributeError(
            # pylint: disable=too-many-format-args
            "{} object has no attribute {}".format(
                self.__class__, attr))

    def __setattr__(self, attr, value):
        """
        Set schema attribute
        """

        # If attribute is in schema, update schema
        if attr in self.__dict__.get("data", {}):
            self.__dict__["data"][attr] = value
        else:  # If not, set attribute on object directly
            self.__dict__[attr] = value

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False # Different types -> not equal

        for field, value in self.data.items():
            if other.data.get(field) != value:
                return False # Different field value

        # All field values are equal
        return True

    def __ne__(self, other):
        """
        Test for not equal and use the inverse of __eq__.
        This is only needed in python2, not python3.
        """
        return not self.__eq__(other)

    def __repr__(self):
        """
        Construnct repr string based on Schema values and optionally
        the serializer field atttribute
        """

        args = []

        for field in self.SCHEMA:
            if field.serializer is False:
                continue

            if self.data.get(field.name) == field.default:
                continue # Exclude default values

            value = self[field.name]

            if field.serializer:
                value = field.serializer(value)

            args.append('{}={!r}'.format(field.name, value))

        return "{}({})".format(self.__class__.__name__, ", ".join(args))



import logging
import sys
import re
import act.api


def setup_logging(loglevel="debug", logfile=None, prefix="act"):
    """setup default logging"""

    numeric_level = getattr(logging, loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: {}'.format(loglevel))

    datefmt = "%Y-%m-%d %H:%M:%S"
    formatter = "[%(asctime)s] app=" + prefix + " level=%(levelname)s msg=%(message)s"

    if logfile:
        logging.basicConfig(
            level=numeric_level,
            filename=logfile,
            format=formatter,
            datefmt=datefmt)
    else:
        logging.basicConfig(
            level=numeric_level,
            stream=sys.stdout,
            format=formatter,
            datefmt=datefmt)


def snake_to_camel(snake):
    """convert snake_case to camelCase"""

    components = snake.split("_")
    return components[0] + "".join([comp.title() for comp in components[1:]])

# https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
def camel_to_snake(camel):
    """convert camelCase to snake_case"""
    camel = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', camel)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', camel).lower()


def prepare_params(params, exclude_params=None, ensure_list=None):
    """convert all keys to camelCase and filter out values. As default,
    act_baseurl and user_id is filtered out"""

    if not exclude_params:
        exclude_params = ["act_baseurl", "user_id", "self"]

    # Put entries if list if they not already contained in a list
    if ensure_list:
        for param in ensure_list:
            if param in params and not isinstance(
                    params[param], (list, tuple)):
                params[param] = [params[param]]

    # Filter out params with null Values and convert to camelCase
    return {
        act.api.utils.snake_to_camel(k): v
        for (k, v) in params.items()
        if (v and k not in exclude_params)}



import json
import copy
import functools
from logging import error, info, debug
import re
import requests
from .schema import Schema, Field, schema_doc, MissingField
from . import RE_UUID_MATCH


class NotImplemented(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)

class InvalidData(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)

class ArgumentError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)

class ResponseError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)

class ValidationError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)

class OriginMismatch(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)

class OriginDoesNotExist(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)

class NotConnected(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


ERROR_HANDLER = {
        # Mapping of message templates provided in 412 errors from backend to
        # Exceptions that will be raised
        "object.not.valid": lambda msg: ValidationError(
            "{message} ({field}={parameter})".format(**msg))
}


def request(method, user_id, url, requests_common_kwargs = None, **kwargs):
    """Perform requests towards API

Args:
    method (str):         POST|GET
    user_id (int):        Act user ID
    url (str):            Absolute URL for the endpoint
    **kwargs (keywords):  Additional options passed to requests json parameter
                          the following fields:
"""

    if not requests_common_kwargs:
        requests_common_kwargs = {}

    requests_kwargs = copy.copy(requests_common_kwargs)

    # Apply common request arguments
    requests_kwargs.update(kwargs)

    # Make sure we have header as argument
    if "headers" not in requests_kwargs:
        requests_kwargs["headers"] = {}

    # Add User ID as header
    requests_kwargs["headers"]["ACT-User-ID"] = str(user_id)

    res = requests.request(
        method,
        url,
        **requests_kwargs
    )

    if res.status_code == 412:
        error_messages = res.json()["messages"]

        # Example output on object validation error:
        # {"responseCode": 412, "limit": 0, "count": 0, "messages": [{"type": "FieldError", "message": "Object did not pass validation against ObjectType.", "messageTemplate": "object.not.valid", "field": "objectValue", "parameter": "127.0.0.x", "timestamp": "2019-09-23T18:19:26.476Z"}], "data": null, "size": 0}

        # Raise ValidationError for 412/Object validation errors
        for msg in error_messages:
            msg_template = msg.get("messageTemplate")
            if msg_template in ERROR_HANDLER:
                raise ERROR_HANDLER[msg_template](msg)

        # All other, unhandled errors - log to error() and raise generic exception
        error("Request failed: {}, {}, {}".format(
            url, kwargs, res.status_code))
        raise ResponseError(res.text)

    elif res.status_code not in (200, 201):
        error("Request failed: {}, {}, {}".format(
            url, kwargs, res.status_code))
        raise ResponseError(
            "Unknown response error {}: {}".format(res.status_code, res.text))

    try:
        return res.json()

    except json.decoder.JSONDecodeError:
        raise ResponseError(
            "Error decoding response {}: {}".format(res.status_code, res.text))


class ActResultSet(object):
    """Represents a list of Act entries"""

    def __init__(self, response, deserializer):
        """Initialize result set
Args:
    response (str):       JSON response from Act. This should include
                          the following fields:
                            - count: the number of entries fetched
                            - limit: the limit in the query
                            - responseCode: responseCode from the API
                            - size: the total number of entries in the platform
                            - data (array): array of entries

    deserializer (class): Deseralizer class
"""

        if not isinstance(response["data"], list):
            raise ResponseError(
                "Response should be list: {}".format(
                    response["data"]))

        self.data = [deserializer(**d) for d in response["data"]]

        self.size = response["size"]
        self.count = response["count"]
        self.limit = response["limit"]
        self.status_code = response["responseCode"]

    @property
    def complete(self):
        """Returns true if we have recieved all data that exists on the endpoint"""
        return self.size >= self.count

    def __call__(self, func, *args, **kwargs):
        """Call function on each data entry"""

        self.data = [getattr(item, func)(*args, **kwargs)
                     for item in self.data]
        return self

    def __len__(self):
        """Returns the number of entries"""
        return len(self.data)

    def __getitem__(self, sliced):
        return self.data[sliced]

    def __str__(self):
        if not self.data:
            return "No result"
        return "\n".join(["{}".format(item) for item in self.data])

    def __bool__(self):
        """ Return True for non empty result sets """
        if self.size > 0:
            return True

        return False

    def __repr__(self):
        """
        Representation of result set
        """

        return repr(self.data)

    def __iter__(self):
        """Iterate over the entries"""
        return self.data.__iter__()


class Config(object):
    """Config object"""
    act_baseurl = None
    user_id = None
    requests_common_kwargs = {}

    def __init__(
            self,
            act_baseurl,
            user_id,
            requests_common_kwargs = None,
            origin_name=None,
            origin_id=None):
        """
        act_baseurl - url to ACT instance
        use_id - ACT user ID
        requests_common_kwargs - options that will be passed to requests when connecting to ACT api
        origin_name - ACT origin name that will be added to all facts where origin is not set
        origin_id - ACT origin id that will be added to all facts where origin is not set

        Only one of origin_name of origin_id must be specified.
        """

        if origin_id and not re.search(RE_UUID_MATCH, origin_id):
            raise ArgumentError("origin_id id not a valaid UUID: {}".format(origin_id))

        self.act_baseurl = act_baseurl
        self.user_id = user_id
        self.requests_common_kwargs = requests_common_kwargs
        self.origin_name = origin_name
        self.origin_id = origin_id


class ActBase(Schema):
    """Act object inheriting Schema, to support serializing and
    deserializing."""

    config = None

    SCHEMA = []

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(ActBase, self).__init__(*args, **kwargs)

    def configure(self, config):
        """Set config object"""

        self.config = config
        return self

    def api_request(self, method, uri, **kwargs):
        """Send request to API and update current object with result"""

        response = request(
            method,
            self.config.user_id,
            "{}/{}".format(self.config.act_baseurl, uri),
            self.config.requests_common_kwargs,
            **kwargs
        )

        return response

    def api_post(self, uri, **kwargs):
        """Send POST request to API with keywords as JSON arguments"""

        return self.api_request("POST", uri, json=kwargs)

    def api_put(self, uri, **kwargs):
        """Send PUT request to API with keywords as JSON arguments"""

        return self.api_request("PUT", uri, json=kwargs)

    def api_delete(self, uri, params=None):
        """Send DELETE request to API
Args:
    uri (str):     URI (relative to base url). E.g. "v1/factType"
    params (Dict): Parameters that are URL enncoded and sent to the API
"""

        return self.api_request("DELETE", uri, params=params)

    def api_get(self, uri, params=None):
        """Send GET request to API
Args:
    uri (str):     URI (relative to base url). E.g. "v1/factType"
    params (Dict): Parameters that are URL enncoded and sent to the API
"""

        return self.api_request("GET", uri, params=params)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False # Different types -> not equal

        for field, value in self.data.items():
            # Only compare serialized fields. The other fields
            # have different representation if they are created locally
            # and not recieved from the back end
            if self.get_field(field).serializer is False:
                continue
            if field == "id":
                # Facts/objects may not have an ID, unless they are returned from the backend
                # We will check for inconsistencies below
                continue
            if other.data.get(field) != value:
                return False # Different field value

        # Two objects where all fields are equal do not have the same id
        if self.id and other.id and self.id != other.id:
            raise InvalidData("Two objects with equal fields do not have the same id")

        # All field values are equal
        return True


class NameSpace(ActBase):
    """Namespace - serialized object specifying Namespace"""

    SCHEMA = [
        Field("name"),
        Field("id"),
    ]


class Organization(ActBase):
    """Manage Organization"""

    SCHEMA = [
        Field("name"),
        Field("id"),
    ]

    def serialize(self):
        # Return None for empty objects (non initialized objects)
        if not self.id:
            return None

        return self.id


@functools.lru_cache(maxsize=128)
def origin_map(config):
    """ Return lookup dictionary (name: uuid) for all known origins """

    # We put this as a separate function, with only config (baseurl, etc)
    # as the parameter. In this way, the cache will be used, since
    # the config will always be the same within the same session

    if not config.act_baseurl:
        raise NotConnected("act_baseurl is not configured, unable to fetch origins")

    # Create base object using the specified configuration
    base = ActBase()
    base.configure(config)

    debug("Looking up origins")

    # Return dictionary of name -> uuid of origins
    return {
        origin["name"]: origin["id"]
        for origin in base.api_get(
            "v1/origin",
            params={"limit": 0, "includeDeleted": False})["data"]
    }


def origin_lookup_serializer(origin, config=None):
    """ Serializer for origins that will lookup by name and verify name/id"""

    if config and config.act_baseurl:  # type: ignore
        # If we have specified act_baseurl, i.e we are connected to a backend,
        # serialize origin to uuid (or None)

        if not origin:
            return None

        if origin.id and not origin.name:
            # Origin specified by uuid, use this directly
            return origin.id

        if origin.name and not origin.id:
            # Lookup uuid by name
            origin_id = origin_map(config).get(origin.name)
            if not origin_id:
                raise OriginDoesNotExist("Unable to find origin with name {}".format(origin.name))
            return origin_id

        if origin.id and origin.name:
            if origin_map(config).get(origin.name) == origin.id:
                return origin.id

            raise OriginMismatch("Origin name and uuid specified, " +
                                 "but the uuid ({}) does not represent ".format(origin.id) +
                                 "the origin with this name ({})".format(origin.name))
        # No origin
        return None

    # Use default serialization, which will include a dictionary of the origin
    # object
    return origin.serialize()


class Origin(ActBase):
    """Manage Origin"""

    SCHEMA = [
        Field("name"),
        Field("id"),
        Field("namespace", serializer=False, deserializer=NameSpace),
        Field("organization", deserializer=Organization),
        Field("description"),
        Field("trust"),
        Field("type", serializer=False),
        Field("flags", serializer=False),
    ]

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(Origin, self).__init__(*args, **kwargs)

    def get(self):
        """Get Origin"""

        if not self.id:
            raise MissingField(
                "Must have fact ID to get origin")

        origin = self.api_get("v1/origin/uuid/{}".format(self.id))["data"]
        self.data = {}
        self.deserialize(**origin)
        return self

    def add(self):
        """Add Origin"""
        params = self.serialize()

        origin = self.api_post("v1/origin", **params)["data"]

        # Empty data and load new result from response
        self.data = {}
        self.deserialize(**origin)

        info("Created origin: {}".format(self.name))

        return self

    def delete(self):
        """Delete Origin"""

        if not self.id:
            raise MissingField(
                "Must have fact ID to delete origin")

        origin = self.api_delete("v1/origin/uuid/{}".format(self.id))["data"]
        self.data = {}
        self.deserialize(**origin)

        info("Deleted origin: {}".format(self.name))
        return self


class Comment(ActBase):
    """Namespace - serialized object specifying Namespace"""

    SCHEMA = [
        Field("comment"),
        Field("id"),
        Field("timestamp", serializer=False),
        Field("reply_to"),
        Field("origin", deserializer=Origin, serializer=False),
    ]



""" Test for act helpers """

import pytest
import act.api


def test_add_uri_fqdn() -> None:  # type: ignore
    """ Test for extraction of facts from uri with fqdn """
    api = act.api.Act("", None, "error")

    uri = "http://www.mnemonic.no/home"

    facts = act.api.helpers.uri_facts(api, uri)

    assert len(facts) == 4
    assert api.fact("componentOf").source("fqdn", "www.mnemonic.no").destination("uri", uri) \
        in facts
    assert api.fact("componentOf").source("path", "/home").destination("uri", uri) in facts
    assert api.fact("scheme", "http").source("uri", uri) in facts
    assert api.fact("basename", "home").source("path", "/home") in facts


def test_uri_should_fail() -> None:  # type: ignore
    """ Test for extraction of facts from uri with ipv4 """
    api = act.api.Act("", None, "error")

    with pytest.raises(act.api.base.ValidationError):
        act.api.helpers.uri_facts(api, "http://")

    with pytest.raises(act.api.base.ValidationError):
        act.api.helpers.uri_facts(api, "www.mnemonic.no")

    with pytest.raises(act.api.base.ValidationError):
        act.api.helpers.uri_facts(api, "127.0.0.1")


def test_add_uri_ipv4() -> None:  # type: ignore
    """ Test for extraction of facts from uri with ipv4 """
    api = act.api.Act("", None, "error")

    uri = "http://127.0.0.1:8080/home"

    facts = act.api.helpers.uri_facts(api, uri)

    assert len(facts) == 5
    assert api.fact("componentOf").source("ipv4", "127.0.0.1").destination("uri", uri) in facts
    assert api.fact("componentOf").source("path", "/home").destination("uri", uri) in facts
    assert api.fact("scheme", "http").source("uri", uri) in facts
    assert api.fact("basename", "home").source("path", "/home") in facts
    assert api.fact("port", "8080").source("uri", uri) in facts


def test_add_uri_ipv6() -> None:  # type: ignore
    """ Test for extraction of facts from uri with ipv4 """
    api = act.api.Act("", None, "error")

    uri = "http://[2001:67c:21e0::16]"

    facts = act.api.helpers.uri_facts(api, uri)

    assert len(facts) == 2
    assert api.fact("scheme", "http").source("uri", uri) in facts
    assert api.fact("componentOf").source("ipv6", "2001:067c:21e0:0000:0000:0000:0000:0016").destination("uri", uri) \
        in facts


def test_add_uri_ipv6_with_port_path_query() -> None:  # type: ignore
    """ Test for extraction of facts from uri with ipv6, path and query """
    api = act.api.Act("", None, "error")

    uri = "http://[2001:67c:21e0::16]:8080/path?q=a"

    facts = act.api.helpers.uri_facts(api, uri)

    assert len(facts) == 6
    assert api.fact("scheme", "http").source("uri", uri) in facts
    assert api.fact("componentOf").source("ipv6", "2001:067c:21e0:0000:0000:0000:0000:0016").destination("uri", uri) \
        in facts
    assert api.fact("port", "8080").source("uri", uri) in facts
    assert api.fact("componentOf").source("path", "/path").destination("uri", uri) in facts
    assert api.fact("basename", "path").source("path", "/path") in facts
    assert api.fact("componentOf").source("query", "q=a").destination("uri", uri) in facts


def test_ip_obj() -> None:
    """ Test ip handling """

    api = act.api.Act("", None, "error")

    assert act.api.helpers.ip_obj("2001:67c:21e0::16") == ("ipv6", "2001:067c:21e0:0000:0000:0000:0000:0016")
    assert act.api.helpers.ip_obj("::1") == ("ipv6", "0000:0000:0000:0000:0000:0000:0000:0001")
    assert act.api.helpers.ip_obj("127.0.0.1") == ("ipv4", "127.0.0.1")

    assert act.api.helpers.ip_obj("127.000.00.01") == ("ipv4", "127.0.0.1")

    with pytest.raises(ValueError):
        assert act.api.helpers.ip_obj("x.y.z") == ("ipv4", "x.y.x")

    with pytest.raises(ValueError):
        assert act.api.helpers.ip_obj("300.300.300.300") == ("ipv4", "x.y.x")

    assert api.fact("resolvesTo") \
            .source("fqdn", "localhost") \
            .destination("ipv4", "127.0.0.1") \
        == api.fact("resolvesTo") \
            .source("fqdn", "localhost") \
            .destination(*act.api.helpers.ip_obj("127.0.0.1"))
